package com.notuvy.thread;

import org.apache.log4j.Logger;

import java.util.concurrent.ThreadFactory;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.HashMap;
import java.util.Map;

/**
 * A ThreadFactory that sets an explicit name for the Thread and also defaults it
 * to be a daemon and sets a simple uncaught exception handler.
 * It also keeps track of all produced threads, so it can check if the current thread
 * originated from this factory.
 *
 * @author murali
 */
public class NamedThreadFactory implements ThreadFactory {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger("com.notuvy.thread");


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final AtomicInteger fIndex = new AtomicInteger(0);
    private final String fName;
    private final boolean fDaemon;
    private final Thread.UncaughtExceptionHandler fHandler;
    private final Map<String,Thread> fThreads = new HashMap<String,Thread>();


    //------------------------------------------------------------
    //- Constructors

    public NamedThreadFactory(String pName, boolean pDaemon, Thread.UncaughtExceptionHandler pHandler) {
        fName = pName;
        fDaemon = pDaemon;
        fHandler = pHandler;
    }

    public NamedThreadFactory(String pName, boolean pDaemon) {
        this (
                pName,
                pDaemon,
                new Thread.UncaughtExceptionHandler() {
                    public void uncaughtException(Thread pThread, Throwable pThrowable) {
                        LOG.error("Fatal thread error.", pThrowable);
                    }
                }
        );
    }

    public NamedThreadFactory(String pName) {
        this (pName, true);
    }


    //------------------------------------------------------------
    //- Gettors


    //------------------------------------------------------------
    //- Settors


    //------------------------------------------------------------
    //- Private/Protected Utility Functions


    //------------------------------------------------------------
    //- Public Interface Functions

    /**
     * Determine whether the current thread originated from this factory.
     *
     * @return a boolean flag indicating that this factory originated the current thread.
     */
    public boolean originatedCurrent() {
        return fThreads.values().contains(Thread.currentThread());
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    public Thread newThread(Runnable pRunnable) {
        String name = String.format("%s-%d", fName, fIndex.incrementAndGet());
        Thread result = new Thread(pRunnable, name);
        result.setDaemon(fDaemon);
        result.setUncaughtExceptionHandler(fHandler);

        fThreads.put(name, result);

        return result;
    }


    //------------------------------------------------------------
    //- Inner Classes


    //------------------------------------------------------------
    //- Main


}
