/*
   Copyright 2008-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.xml;

import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import java.io.*;
import java.util.Collections;
import java.util.List;

/**
 * A bean that has an XML representation.
 *
 * User: murali
 * Date: Jan 29, 2008
 * Time: 6:17:08 AM
 */
public abstract class SimpleJdom {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger("com.notuvy.xml");

    private static final List<Element> EMPTY_LIST = Collections.emptyList();

    private static final SimpleJdom DUMMY = new SimpleJdom() {
        public Element toXml() {
            return null;
        }

        protected void parse(Element pRoot) {
        }

        @Override
        public String toString() {
            return "DUMMY";
        }
    };

    private static final XmlStorable NO_STORAGE = new BaseStorage(DUMMY) {
        public XmlIOStatus<Void> save() {
            return XmlIOStatus.success();
        }

        public XmlIOStatus<Void> read() {
            return XmlIOStatus.success();
        }

        public XmlIOStatus<Void> readIfStoredImageExists() {
            return XmlIOStatus.success();
        }

        public XmlIOStatus<Void> readOrCreate() {
            return XmlIOStatus.success();
        }

        public XmlIOStatus<Void> delete() {
            return XmlIOStatus.success();
        }

        public XmlIOStatus<Boolean> exists() {
            return new XmlIOStatus<Boolean>(true);
        }

        @Override
        public String toString() {
            return "NO STORAGE";
        }
    };

    private static final Format XML_OUTPUT_FORMAT = Format.getRawFormat();

    static {
        XML_OUTPUT_FORMAT.setIndent("  ");
    }


    //------------------------------------------------------------
    //- Class Functions

    public static Element addChildElement(Element pParent, String pTagName) {
        Element result = new Element(pTagName);
        pParent.addContent(result);
        return result;
    }

    public static List<Element> getChildren(Element pParent, String pChildTag) {
        @SuppressWarnings("unchecked")
        List<Element> result = (List<Element>) pParent.getChildren(pChildTag);
        if (result == null) {
            result = EMPTY_LIST;
        }
        return result;
    }

    public static List<Element> getChildrenOfContainer(Element pParent, String pChildTag, String pContainerTag) {
        List<Element> result = EMPTY_LIST;
        Element container = pParent.getChild(pContainerTag);
        if (container != null) {
            result = getChildren(container, pChildTag);
        }
        return result;
    }

    public static String getChildText(Element pParent, String pChildTag) {
        String result = null;  // Pessimistic.
        Element element = pParent.getChild(pChildTag);
        if (element != null) {
            result = element.getText();
        }
        return result;
    }

    public static Long getChildLong(Element pParent, String pChildTag) {
        String string = getChildText(pParent, pChildTag);
        return (string == null) ? null : Long.parseLong(string);
    }

    public static Integer getChildInteger(Element pParent, String pChildTag) {
        String string = getChildText(pParent, pChildTag);
        return (string == null) ? null : Integer.parseInt(string);
    }


    //------------------------------------------------------------
    //- Instance Variables

    private Format fFormat = null;
    private transient XmlStorable fStorage = NO_STORAGE;


    //------------------------------------------------------------
    //- Constructors

    protected SimpleJdom() {
    }

    protected SimpleJdom(String pContent) {
        readFromContent(pContent);
    }

    //------------------------------------------------------------
    //- Accessors

    protected Format getFormat() {
        if (fFormat == null) {
            fFormat = createFormat();
        }

        return fFormat;
    }

    public XmlStorable getStorage() {
        return fStorage;
    }


    //------------------------------------------------------------
    //- Settors

    public void setStorage(XmlStorable pStorage) {
        fStorage = pStorage;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    protected boolean isValidRoot(Element pRoot) {
        boolean result = rootName().equals(pRoot.getName());

        if (!result) {
            LOG.error("Unexpected root: " + pRoot.getName());
        }

        return result;
    }

    //------------------------------------------------------------
    //- Public Interface Functions

    public XmlIOStatus<Void> saveTo(Writer pWriter) {
        XmlIOStatus<Void> success = XmlIOStatus.failure();  // Pessimistic.

        XMLOutputter serializer = new XMLOutputter(getFormat());
        try {
            serializer.output(toDocument(), pWriter);
            success = XmlIOStatus.success();
        } catch (IOException ioe) {
            LOG.error("Problem serializing XML to output.", ioe);
        }

        return success;
    }

    public XmlIOStatus<Void> readFrom(InputStream pInputStream) {
        InputStreamReader reader = new InputStreamReader(pInputStream);
        XmlIOStatus<Void> success = readFrom(reader);
        try {
            reader.close();
        } catch (IOException ioe) {
            LOG.error("XML data read problem.", ioe);
            success = XmlIOStatus.failure("XML data read problem.");
        }
        return success;
    }

    public XmlIOStatus<Void> readFrom(Reader pReader) {
        XmlIOStatus<Void> status = XmlIOStatus.failure();  // Pessimistic.
        SAXBuilder builder = new SAXBuilder();
        try {
            Document document = builder.build(pReader);
            parse(document.getRootElement());
            status = XmlIOStatus.success();
        } catch (JDOMException jdome) {
            reportXmlParseError(jdome);
        } catch (IOException ioe) {
            LOG.error("XML data read problem.", ioe);
        }
        return status;
    }

    public XmlIOStatus<Void> readFromContent(String pString) {
        StringReader reader = new StringReader(pString);
        XmlIOStatus<Void> success = readFrom(reader);
        reader.close();
        return success;
    }

    public String toXmlString() {
        StringWriter writer = new StringWriter();
        saveTo(writer);
        try {
            writer.close();
        } catch (IOException ioe) {
            LOG.error("Problem serializing to string.", ioe);
        }
        return writer.toString();
    }

    public Element createRoot() {
        return new Element(rootName());
    }

    public Document toDocument() {
        Document result = new Document();
        result.setRootElement(toXml());
        return result;
    }


    // TODO: public void populateFrom(SimpleJdom pOther), comparing rootName();


    //------------------------------------------------------------
    //- Class Interface Functions

    /**
     * Return this object as a sub-element to be inserted within another object.
     * This does NOT return a Document representation.
     *
     * @return Element representation suitable for insertion into another object.
     */
    public abstract Element toXml();

    protected abstract void parse(Element pRoot);

    public String rootName() {
        throw new RuntimeException("No root name defined.");
    }

    protected Format createFormat() {
        return XML_OUTPUT_FORMAT;
    }

    public String toString() {
        return toXmlString();
    }

    protected void reportXmlParseError(JDOMException jdome) {
        LOG.error("XML structure problem", jdome);
    }


    //------------------------------------------------------------
    //- Inner Classes

    //------------------------------------------------------------
    //- Main
}
